
/* ==================================================================== */
/*			MIDI dump module				*/
/* ==================================================================== */

/* Included files */

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <termios.h>
#include <linux/link.h>
#include <asm/io.h>

#include "../proteus.h"
#include "../canvas.h"
#include "../fm_def.h"
#include "mpu401.h"

extern char fname[];
extern _paramblk Params;
extern unsigned char playstat[], scoreorder[];
extern int Link;
extern int miditype;
extern unsigned char *Tracks[];
extern unsigned char *Tkptr[];

#undef  RSTATUS

#define uRESOL	500	/* usec */

unsigned char MThd[] =
	{ 0x4d,0x54,0x68,0x64,	/* offset 0 into file */
	  0,0,0,6,
	  0,1,
	  0,33,
	  4,0			/* resol=500us */
	};

unsigned char MTrk0[] =
	{ 0x4d,0x54,0x72,0x6b,
	  0,0,0,26,
/* --------------------------------------------------------------------
FF 58 04 nn dd cc bb	Time Signature 
The time signature is expressed as four numbers.  nn and dd represent 
the numerator and denominator of the time signature as it would be 
notated.  The denominator is a negative power of two:  2 represents a 
quarter-note, 3 represents an eighth-note, etc.  The cc parameter 
expresses the number of MIDI clocks in a metronome click.  The bb 
parameter expresses the number of notated 32nd-notes in a MIDI quarter-
note (24 MIDI Clocks).  This was added because there are already 
multiple programs which allow the user to specify that what MIDI thinks 
of as a quarter-note (24 clocks) is to be notated as, or related to in 
terms of, something else. 

Therefore, the complete event for 6/8 time, where the metronome clicks 
every three eighth-notes, but there are 24 clocks per quarter-note, 72 
to the bar, would be (in hex):

	FF 58 04 06 03 24 08

That is, 6/8 time (8 is 2 to the 3rd power, so this is 06 03), 32 MIDI 
clocks per dotted-quarter (24 hex!), and eight notated 32nd-notes per 
MIDI quarter note.
-------------------------------------------------------------------- */
	  0,	0xff,0x58,4,	4,2,24,8,
/* @q=120 5E5us = 0x7a120 */
/* --------------------------------------------------------------------
FF 51 03 tttttt    	Set Tempo, in microseconds per MIDI quarter-note
This event indicates a tempo change.  Another way of putting 
"microseconds per quarter-note" is "24ths of a microsecond per MIDI 
clock".  Representing tempos as time per beat instead of beat per time 
allows absolutely exact long-term synchronization with a time-based sync 
protocol such as SMPTE time code or MIDI time code.  This amount of 
accuracy provided by this tempo resolution allows a four-minute piece at 
120 beats per minute to be accurate within 500 usec at the end of the 
piece.  Ideally, these events should only occur where MIDI clocks would 
be located Q this convention is intended to guarantee, or at least 
increase the likelihood, of compatibility with other synchronization 
devices so that a time signature/tempo map stored in this format may 
easily be transferred to another device. 

-------------------------------------------------------------------- */
	  0,	0xff,0x51,0x03,	0x07,0xa1,0x20,
/* --------------------------------------------------------------------
FF 59 02 sf mi	Key Signature
        sf = -7:  7 flats
        sf = -1:  1 flat
        sf = 0:  key of C
        sf = 1:  1 sharp
        sf = 7: 7 sharps
        
        mi = 0:  major key
        mi = 1:  minor key
-------------------------------------------------------------------- */
	  0,	0xff,0x59,0x02,	-3,1,
	  0x83,0,0xff,0x2f,0
	};

/* --------------------------------------------------------------------
FF 7F len data	Sequencer-Specific Meta-Event

        Special requirements for particular sequencers may use this 
event type:  the first byte or bytes of data is a manufacturer ID.  
However, as this is an interchange format, growth of the spec proper is 
preferred to use of this event type.  This type of event may be used by 
a sequencer which elects to use this as its only file format;  
sequencers with their established feature-specific formats should 
probably stick to the standard features when using this format.
-------------------------------------------------------------------- */
	
unsigned long MTrk   = 0x6b72544d;
unsigned long EOtrk  = 0x002fff00;
unsigned long AllOff = 0x007bb000;

/* ==================================================================== */


static int t2var(unsigned long t,FILE *fd)
{
  int n=1;
  unsigned char c;
#ifdef MDEBUG
	fprintf(stderr,"                 t2var of %8ld =",t);
#endif
	if (t&0xfe00000) {
		c = ((t>>21)&0x7f)|0x80;
#ifdef MDEBUG
		fprintf(stderr," %02x",c);
#endif
		fputc(c,fd); n++;
	}
	if (t&0x1fc000)  {
		c = ((t>>14)&0x7f)|0x80;
#ifdef MDEBUG
		fprintf(stderr," %02x",c);
#endif
		fputc(c,fd); n++;
	}
	if (t&0x3f80)    {
		c = ((t>>7)&0x7f)|0x80;
#ifdef MDEBUG
		fprintf(stderr," %02x",c);
#endif
		fputc(c,fd);  n++;
	}
	c = t&0x7f;
#ifdef MDEBUG
	fprintf(stderr," %02x\n",c);
#endif
	fputc(c,fd);
	return(n);
}

static int v2big(unsigned long v,FILE *fd)
{
	fputc((v>>24),fd); fputc((v>>16),fd);
	fputc((v>>8),fd); fputc(v,fd);
	return(4);
}


static void n2f(unsigned n,FILE *fd)
{
  char nT1[12] = "CCDEEFFGGABB";
  char nT2[12] = " # b  # # b ";
  int nn,no;
	nn=n%12; no=(n/12)-2;
	fprintf(fd,"%c%c%d",nT1[nn],nT2[nn],no);
	if (no>=0) fprintf(fd," ");
}


static void dumppar(_paramblk *p,FILE *fd)
{
 int i,j,k,ns;
 Instr *I;
 
  fprintf(fd,"# Function\t%s\n",fname);
  fprintf(fd,"# Params\n");
  fprintf(fd,"#  F[0]=%12.6f\tF[1]=%12.6f\n",p->fp[0],p->fp[1]);
  fprintf(fd,"#  F[2]=%12.6f\tF[3]=%12.6f\n",p->fp[2],p->fp[3]);
  fprintf(fd,"#  I[0]=%12d\tI[1]=%12d\n",p->ip[0],p->ip[1]);
  fprintf(fd,"#  I[2]=%12d\tI[3]=%12d\n",p->ip[2],p->ip[3]);
  fprintf(fd,"#  X[0]=%12d\tX[1]=%12d\n",p->xp[0],p->xp[1]);
  fprintf(fd,"#  X[2]=%12d\tX[3]=%12d\n",p->xp[2],p->xp[3]);
  fprintf(fd,"#  X[4]=%12d\tX[5]=%12d\n",p->xp[4],p->xp[5]);
  fprintf(fd,"#  X[6]=%12d\tX[7]=%12d\n",p->xp[6],p->xp[7]);
  for (ns=0; ns<VOICES; ns++) {
  	i = scoreorder[ns];
  	I = &(Params.instr[i]);
  	fprintf(fd,"#  Ins[%2d] %-20s",i,I->name);
  	fprintf(fd," %2d %2d %3d %3d %02d %2d %2d\n",I->mpu,I->chan,I->patch,
  		I->volume,I->flags,I->rmin,I->rmax);
  }
  fprintf(fd,"#  M\n");
  k=0;
  for (i=0; i<11; i++) {
  	fprintf(fd,"#  ");
  	for (j=0; j<12; j++)
  		if (k<128) fprintf(fd,"%c",Params.Mtempl[k++]+'0');
  	fprintf(fd,"\n");
  }
  fprintf(fd,"#\n");
}


/* ==================================================================== */


void trkdump(char *fname)
{
  int i,nsco,ntrk,firstev,l,ffmrk;
  unsigned long t,t0,t1,dt;
  FILE *fd;
  double g,dg;
  unsigned char *p,*Ps,*s,*s1;

	remove(fname);
	fd=fopen(fname,"w");
	dumppar(&Params,fd);

	fprintf(fd,"Track 00(FF)\n");
	p = Tracks[0];
	firstev = 1;
	while (p < Tkptr[0]) {
		t1 = *((unsigned long *)p); p+=4;
		l = *p++;

		if (*p==0xff) {
		  	if ((p[1]==1) || (p[1]==6)) {
				if (firstev) { t0=dt=t1; firstev=0; }
				  else { dt=t1-t0; t0=t1; }
				g = t1/1000000.0;
				dg = dt/1000000.0;
		  	}
		  	if (p[1]==1) {
/* --------------------------------------------------------------------
FF 01 len text	Text Event
Any amount of text describing anything.  It is a good idea to put a text 
event right at the beginning of a track, with the name of the track, a 
description of its intended orchestration, and any other information 
which the user wants to put there.  Text events may also occur at other 
times in a track, to be used as lyrics, or descriptions of cue points.  
The text in this event should be printable ASCII characters for maximum 
interchange.  However, other character codes using the high-order bit 
may be used for interchange of files between different programs on the 
same computer which supports an extended character set.  Programs on a 
computer which does not support non-ASCII characters should ignore those 
characters.
-------------------------------------------------------------------- */
	  		  l=p[2]; p+=3;
			  fprintf(fd,"%12.6f %12.6f  ",g,dg);
	  		  fprintf(fd,"COMMENT \"");
	  		  while (l--) fprintf(fd,"%c",*p++);
	  		  fprintf(fd,"\"\n");
	  		}
	  		else if (p[1]==6) {
	  		 int u;
/* --------------------------------------------------------------------
FF 06 len text	Marker
Normally in a format 0 track, or the first track in a format 1 file.  
The name of that point in the sequence, such as a rehearsal letter or 
section name ("First Verse", etc.). 
-------------------------------------------------------------------- */
	  		  l=p[2]; p+=3;
			  fprintf(fd,"%12.6f %12.6f  ",g,dg);
	  		  fprintf(fd,"MARKER");
	  		  if (l==8) {
	  		    memcpy(&u,p,4); fprintf(fd," %8d",u);
	  		    memcpy(&u,p+4,4); fprintf(fd," %8d",u);
	  		    p+=8; l=0;
	  		  }
	  		  else if (l==12) {
	  		    char buf[16];
	  		    memcpy(buf,p,8); buf[7]=0;
	  		    memcpy(&u,p+8,4); fprintf(fd," %-8s %8d",buf,u);
	  		    p+=8; l=0;
	  		  }
	  		  else {
	  		    while (l--) fprintf(fd," %02x",*p++);
			  }
			  fprintf(fd,"\n");
	  		}
		}
	}
	fprintf(fd,"\n");

	for (nsco=0; nsco<VOICES; nsco++) {
		ntrk = scoreorder[nsco];
		if (Tkptr[ntrk]==Tracks[ntrk]) continue;

		fprintf(fd,"Track %02d(%02d)\t%s\n",nsco+1,ntrk,
			Params.instr[ntrk].name);
		p = Tracks[ntrk];
		Ps = playstat+ntrk*128;
		memset(Ps,0,128);
		firstev = 1;
		while (p < Tkptr[ntrk]) {
			t1 = *((unsigned long *)p); p+=4;
			l = *p++;
			s = Ps+p[1];
			ffmrk = ((*p==0xff) && ((p[1]==1) || (p[1]==6)));

			if (!ffmrk) {
			  g = t1/1000000.0;
			  if (firstev) { t0=dt=t1; firstev=0; }
			    else { dt=t1-t0; t0=t1; }
			  dg = dt/1000000.0;
			  fprintf(fd,"%12.6f %12.6f  ",g,dg);
			}

			switch(*p&0xf0) {
			  case 0x80:	if (*s) {
			  		  *s=0; fprintf(fd,"OFF ");
			  		}
			  		else fprintf(fd,"OFF-err ");
			  		n2f(p[1],fd);
			  		fprintf(fd," %d",p[2]);
			  		p+=3;
			  		break;
			  case 0x90:	if (*s==0) {
			  		  *s=1; fprintf(fd,"ON  ");
			  		}
			  		else fprintf(fd,"ON-err ");
			  		n2f(p[1],fd);
			  		fprintf(fd," %d",p[2]);
			  		p+=3;
			  		break;

			  case 0xb0:	if (p[1]==0x7b) {
					  s1=Ps;
					  fprintf(fd,"OFF>");
					  for (i=0; i<128; i++) {
					    if (s1[i]) {
			  		  	s1[i]=0; n2f(i,fd);
			  		  	fprintf(fd," ");
					    }
					  }
					  p+=3;
			  		}
			  		else {
					  while (l--) fprintf(fd,"%02x ",*p++);
					}
					break;

			  case 0xf0:    if (ffmrk) {
			  		    p+=l; break;
			  		}
			  		/* fall thru */
			  default:
			  	while (l--) fprintf(fd,"%02x ",*p++);
			}
			if (!ffmrk) fprintf(fd,"\n");
		}
		s=Ps;
		for (i=0; i<128; i++) {
		  if (s[i]) {
		    fprintf(fd,"%12.6f %12.6f *OFF ",g,0.0);
		    n2f(i,fd);
		    fprintf(fd,"\n");
		  }
		}
		fprintf(fd,"\n");
	}
	fclose(fd);
}


/* ==================================================================== */


void mididump(char *fname)
{
  int i,ntrk,firstev,l,valid,nvalid;
  int rstat;			/* running status */
  int thischan,thismpu;
  unsigned long t,t0,t1,dt,nb;
  unsigned qresol;
  unsigned long fposz,fptk0;
  FILE *fd;
  unsigned char *p,*Ps,*s,*s1;
  char buf[64];

	remove(fname);
	fd=fopen(fname,"w+");
	fwrite(MThd,sizeof(MThd),1,fd);

/* Track 0: time signature/tempo */
	fptk0 = ftell(fd);
	fwrite(MTrk0,sizeof(MTrk0),1,fd);
	fflush(fd);

	t=MTrk0[22]|(MTrk0[21]<<8)|(MTrk0[20]<<16);
	qresol = t/uRESOL;
#ifdef MDEBUG
	fprintf(stderr,"TRK00  t=%08lx(%ld)  qresol=%d\n\n",t, t, qresol);
#endif

/* Rest of tracks */		
	memset(playstat,0,4096);
	nvalid = 1;

	for (ntrk=0; ntrk<VOICES; ntrk++) {
		thischan = ntrk&15; thismpu = (ntrk>>4)&3;
		Ps = playstat+ntrk*128;
		fposz = ftell(fd);

/* MTrk 0 0 0 0 */
		fwrite(&MTrk,4,1,fd);
		nb = 0;
		fwrite(&nb,4,1,fd);

/* comment at track header */
		sprintf(buf,"%d:%03d %s",Params.instr[ntrk].mpu,
			Params.instr[ntrk].patch,Params.instr[ntrk].name); 
#ifdef MDEBUG
		fprintf(stderr,"TRK%2d  %s %p-%p Ps=%p\n",ntrk+1, buf,
			Tracks[ntrk], Tkptr[ntrk], Ps);
#endif
		s=buf; i=strlen(s);
		nb += (i+4);
/* 0 ff 3 len <string> */
		fputc(0,fd);
		fputc(0xff,fd); fputc(3,fd); fputc(i,fd);
		while (i--) fputc(*s++,fd);

		p = Tracks[ntrk];
		firstev = 1; valid = 0;
		rstat = 0;

		while (p < Tkptr[ntrk]) {
			t1 = *((unsigned long *)p); p+=4;
			if (firstev) { t0=dt=t1; firstev=0; }
			l = *p++;
			s = Ps+p[1];
#ifdef MDEBUG
			fprintf(stderr,"       %08ld: len=%d *Ps=%02x op=%02x\n",
				t1, l, *s, *p);
#endif
			if (*p>=0xf0) rstat=0;

			switch(*p&0xf0) {
			  case 0x80:	if (*s==0) { p+=3; break; }
			  		*s=0; l=3;
			  		goto xmid01;
			  case 0x90:	if (*s) { p+=3; break; }
			  		if (!valid) valid=1;
			  		*s=1; l=3;
			  		goto xmid01;
			  case 0xb0:	if (p[1]==0x7b) {
			  		 int f=1;
			  		  rstat=0; s1=Ps;
					  for (i=0; i<128; i++) {
					    if (s1[i]) {
					      s1[i]=0;
					      if (f) {
					        f=0;
					  	nb += t2var(dt/uRESOL,fd);
					  	dt=t1-t0; t0=t1;
					      }
					      else {
					        fputc(0,fd); nb++;
					      }
					      fputc(0x80|thischan,fd);
					      fputc(i,fd); fputc(0,fd);
					      nb+=3;
					    }
					  }
					  p+=l; break;
			  		}
xmid01:
			  default:	if (*p==0xff) {
			  		  if (p[1]==0x51) {
			  		    memcpy(MTrk0+20,p+3,3);
			  		    t=p[5]|(p[4]<<8)|(p[3]<<16);
			  		    qresol = t/uRESOL;
			  		  }
			  		}
			  		else {
					  nb += t2var(dt/uRESOL,fd);
					  dt=t1-t0; t0=t1;
			  		  if (rstat==*p)
			  		   { fwrite(p+1,1,l-1,fd); nb+=(l-1); }
			  		  else
			  		   { fwrite(p,1,l,fd);
#ifdef RSTATUS
			  		     rstat=*p;
#endif
			  		     nb+=l; }
			  		}
			  		p+=l;
			  		break;
			 } 
		}

		if (valid) {
			nvalid++;
			
			s=Ps;
			for (i=0; i<128; i++) {
			  if (s[i]) {
#ifdef MDEBUG
			    fprintf(stderr,"            01 %02x %02x 00\n",
				0x80|thischan, i);
#endif
			    fputc(1,fd); fputc(0x80|thischan,fd);
			    fputc(i,fd); fputc(0,fd);
			    nb+=4;
			  }
			}
			fwrite(&EOtrk,4,1,fd); nb+=4;
			fflush(fd);
			fseek(fd,fposz+4,SEEK_SET);
			v2big(nb,fd);
			fflush(fd);
			fseek(fd,0,SEEK_END);
#ifdef MDEBUG
			fprintf(stderr,"EOT%2d len=%d\n\n",ntrk+1, nb);
#endif
		}
		else {
			fseek(fd,fposz,SEEK_SET);
			fflush(fd);
		}
	}
	fseek(fd,11,SEEK_SET);
	fwrite(&nvalid,1,1,fd);
	i=qresol>>8; fwrite(&i,1,1,fd);
	fwrite(&qresol,1,1,fd);
	fflush(fd);
	fseek(fd,fptk0,SEEK_SET);
	fwrite(MTrk0,sizeof(MTrk0),1,fd);
	
	if (!valid)
		ftruncate(fileno(fd),fposz);
	fclose(fd);
}


/* ==================================================================== */


void rawdump (char *f1, char *f2)
{
  NoteEvent N;
  FILE *fs, *fd;

  fs = fopen(f1,"r"); if (fs==NULL) return;
  fd = fopen(f2,"w");  

  while (fread (&N, sizeof(NoteEvent), 1, fs) > 0) {
    fprintf(fd,"** a=%08x b=%08x c=%08x lck=%d\n", N.a, N.b, N.c, N.lock);
    fprintf(fd,"   pitch=%d  intensity=%d  timbre=%d\n", N.pitch,
    	N.intensity, N.timbre);
    fprintf(fd,"   dur=%d  deltat=%d\n", N.duration, N.deltat);
    fprintf(fd,"   vol=%d(%d)  press=%d(%d)\n", N.startvol, N.deltavol,
    	N.startpress, N.deltapress);
  }
  
  fclose(fd);
  fclose(fs);
}


/* ==================================================================== */

